import numpy as np
import laspy
import json
import math
import utm # src: https://github.com/Turbo87/utm
from itertools import groupby
from server.data_processing.area_processing import calculate_corners, define_gridareas


# hard coded files for test data
lazData_path = ["server/example_lidar_data/ot_N_000005_1.laz", "server/example_lidar_data/ot_N_000033_1.laz"]

# Info about data
def about_laz_file():
    with laspy.open(lazData_path[0]) as fh:
        # Print metadata properties
        print("File Version:", fh.header.version)
        print("Point Count:", fh.header.point_count)
        print("Scale Factors:", fh.header.scale)
        print("Offset:", fh.header.offset)

        las = fh.read()
        print(las)
        print('Points from data:', len(las.points))
        ground_pts = las.classification == 2
        bins, counts = np.unique(las.return_number[ground_pts], return_counts=True)
        print('Ground Point Return Number distribution:')
        for r, c in zip(bins, counts):
            print('    {}:{}'.format(r, c))

    return [las.header.version, las.header.point_count, las.header.scale, las.header.offset]

# check if lidar points is within range of the area selected
def inArea(position, areaRange):
    y, x, _ = position # position to be checked
    if (abs(areaRange[0][0]) > x > abs(areaRange[1][0])) and (abs(areaRange[0][1]) > abs(y) > abs(areaRange[1][1])):
        return True
    else:
        return False

# find distance between two points
def distance(point1, point2):
    y1, x1 = point1
    y2, x2 = point2
    return math.sqrt(abs(y2 - y1)**2 + abs(x2 - x1)**2)

# find the closest point in json list
def closest_points(point, list, coords_taken):
    closest_point_found = None
    closest_dist = float('inf')
    for current_point in list:
        dist = distance(point, current_point['properties']['sub_div_center'][0])
        if dist < closest_dist and current_point not in coords_taken:
            closest_dist = dist
            closest_point_found = current_point
    return closest_point_found

# Calculation of height in an area from las files
def find_height(points):
    print(points, " 5 ")
    height_differences = [] # final container for height data

    # sort the points
    sorted_coords = sorted(points, key=lambda coord: (coord[0], coord[1]))

    # group the sorted points that has the same xy- coordinates together
    groupCoords = [list(group) for key, group in groupby(sorted_coords, key=lambda coord: (coord[0], coord[1]))]

    # loop through the groups to find the difference in height
    for group in groupCoords:
        if len(group) < 2 or group[0] == group[1]:
            continue # jump over iteration if there is only one coordinate or lidar registered the same point twice

        # find max and min height
        min_height = min(coords[2] for coords in group)
        max_height = max(coords[2] for coords in group)

        # difference between them
        difference = max_height - min_height
        height_differences.append(difference)

    # list of thickness in an area
    return height_differences

# find the height of an area based on the coordinates of it's center
# and it's affiliations (subId and groupId) (soon to be implemented
def calculate_area_data(center, body_of_water):
    # container for all the heights in area
    area_heights = []

    # zone coords taken
    taken_coords = []

    # read json data with path to data for specific water body
    file_name = "server/lake_relations/" + body_of_water + "_div.json"
    with open(file_name) as data:
        map_data = json.load(data)

    # grid cell offset
    cell_x, cell_y = map_data['tile_width'], map_data['tile_height']
    # convert the offset to meter
    cell_x = 111.320 * cell_x
    cell_y = 111.320 * math.cos(60) * cell_y

    # set the limit of the area compared to local coordinates
    area_limit = calculate_corners(center[0], center[1], (cell_x, cell_y))

    # grid data
    map_data = map_data['features']

    print(map_data[0]['properties']['sub_div_center'])

    map_zones = [area_limit[1],area_limit[3]]

    map_data = list(filter(lambda point_position: inArea((point_position['properties']['sub_div_center'][0], point_position['properties']['sub_div_center'][1],0.0), map_zones), map_data))

    # Refactor lidar data to a readable format
    iceOver = laspy.read(lazData_path[0])
    iceUnder = laspy.read(lazData_path[1])

    # add two files together temporary test data(soon to be removed)
    ice_points = list(zip(iceOver.X,iceOver.Y,iceOver.Z)) + list(zip(iceUnder.X,iceUnder.Y,iceUnder.Z))
    max_point = max(ice_points)
    min_point = min(ice_points)

    # define all the sub-areas within the area, local coordinates
    grid_area_heights = define_gridareas(60,10, (cell_x, cell_y),4)

    # find the heights of each sub-area => area-heights
    for sub_area in grid_area_heights:
        print(sub_area)
        start = min(sub_area[1])
        end = max(sub_area[1])

        #test data
        # zone coordinates sett to be relative to the lidar data
        areazone = [(((min_point[0] - max_point[0]) * ((start[0]-center[0])/(area_limit[1][0]-area_limit[3][0]))) + (min_point[0] - max_point[0])/2 + max_point[0],
                     (((min_point[1] - max_point[1]) * ((start[1]-center[1])/(area_limit[1][1]-area_limit[3][1]))) + (min_point[1] - max_point[1])/2 + min_point[1])),
                    ((((min_point[0] - max_point[0]) * ((end[0]-center[0])/(area_limit[1][0]-area_limit[3][0]))) + (min_point[0] - max_point[0])/2 + max_point[0],
                     (((min_point[1] - max_point[1]) * ((end[1]-center[1])/(area_limit[1][1]-area_limit[3][1])))) + (min_point[1] - max_point[1])/2 + min_point[1]))]

        # calculate map zones height
        ys, xs = start
        ye, xe = end
        sub_center = ((xs + xe)/2, (ys + ye)/2)
        if len(map_data) > 0:
            current_map_zone = closest_points(sub_center, map_data, taken_coords)
            taken_coords.append((current_map_zone['properties']['sub_div_center'][0][0], current_map_zone['properties']['sub_div_center'][0][1]))
        else:
            return [0]

        current_zone_id = current_map_zone['properties']['sub_div_id']

        # filter data within sub-area zones
        points_in_area = list(filter(lambda point_position: inArea(point_position, areazone), ice_points))
        heights = find_height(points_in_area)
        area_heights.append((current_zone_id, sub_area[0], sub_center, heights))

    return area_heights

print(calculate_area_data((61,11), 0.04,'mjosa'))

